# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/onprem',
                'doc_host': 'https://amaiya.github.io',
                'git_url': 'https://github.com/amaiya/onprem',
                'lib_path': 'onprem'},
  'syms': { 'onprem.console': {},
            'onprem.core': { 'onprem.core.AnswerConversationBufferMemory': ('core.html#answerconversationbuffermemory', 'onprem/core.py'),
                             'onprem.core.AnswerConversationBufferMemory.save_context': ( 'core.html#answerconversationbuffermemory.save_context',
                                                                                          'onprem/core.py'),
                             'onprem.core.LLM': ('core.html#llm', 'onprem/core.py'),
                             'onprem.core.LLM.__init__': ('core.html#llm.__init__', 'onprem/core.py'),
                             'onprem.core.LLM.ask': ('core.html#llm.ask', 'onprem/core.py'),
                             'onprem.core.LLM.chat': ('core.html#llm.chat', 'onprem/core.py'),
                             'onprem.core.LLM.check_model': ('core.html#llm.check_model', 'onprem/core.py'),
                             'onprem.core.LLM.download_model': ('core.html#llm.download_model', 'onprem/core.py'),
                             'onprem.core.LLM.ingest': ('core.html#llm.ingest', 'onprem/core.py'),
                             'onprem.core.LLM.load_chatqa': ('core.html#llm.load_chatqa', 'onprem/core.py'),
                             'onprem.core.LLM.load_ingester': ('core.html#llm.load_ingester', 'onprem/core.py'),
                             'onprem.core.LLM.load_llm': ('core.html#llm.load_llm', 'onprem/core.py'),
                             'onprem.core.LLM.load_qa': ('core.html#llm.load_qa', 'onprem/core.py'),
                             'onprem.core.LLM.load_vectordb': ('core.html#llm.load_vectordb', 'onprem/core.py'),
                             'onprem.core.LLM.prompt': ('core.html#llm.prompt', 'onprem/core.py')},
            'onprem.guider': { 'onprem.guider.Guider': ('guider.html#guider', 'onprem/guider.py'),
                               'onprem.guider.Guider.__init__': ('guider.html#guider.__init__', 'onprem/guider.py'),
                               'onprem.guider.Guider.prompt': ('guider.html#guider.prompt', 'onprem/guider.py')},
            'onprem.ingest': { 'onprem.ingest.Ingester': ('ingest.html#ingester', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.__init__': ('ingest.html#ingester.__init__', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.get_db': ('ingest.html#ingester.get_db', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.get_embedding_model': ( 'ingest.html#ingester.get_embedding_model',
                                                                               'onprem/ingest.py'),
                               'onprem.ingest.Ingester.ingest': ('ingest.html#ingester.ingest', 'onprem/ingest.py'),
                               'onprem.ingest.MyElmLoader': ('ingest.html#myelmloader', 'onprem/ingest.py'),
                               'onprem.ingest.MyElmLoader.load': ('ingest.html#myelmloader.load', 'onprem/ingest.py'),
                               'onprem.ingest.batchify_chunks': ('ingest.html#batchify_chunks', 'onprem/ingest.py'),
                               'onprem.ingest.does_vectorstore_exist': ('ingest.html#does_vectorstore_exist', 'onprem/ingest.py'),
                               'onprem.ingest.load_documents': ('ingest.html#load_documents', 'onprem/ingest.py'),
                               'onprem.ingest.load_single_document': ('ingest.html#load_single_document', 'onprem/ingest.py'),
                               'onprem.ingest.process_documents': ('ingest.html#process_documents', 'onprem/ingest.py')},
            'onprem.pipelines': { 'onprem.pipelines.Summarizer': ('pipelines.html#summarizer', 'onprem/pipelines.py'),
                                  'onprem.pipelines.Summarizer.__init__': ('pipelines.html#summarizer.__init__', 'onprem/pipelines.py'),
                                  'onprem.pipelines.Summarizer._map_reduce': ( 'pipelines.html#summarizer._map_reduce',
                                                                               'onprem/pipelines.py'),
                                  'onprem.pipelines.Summarizer._refine': ('pipelines.html#summarizer._refine', 'onprem/pipelines.py'),
                                  'onprem.pipelines.Summarizer.summarize': ('pipelines.html#summarizer.summarize', 'onprem/pipelines.py')},
            'onprem.utils': { 'onprem.utils.download': ('utils.html#download', 'onprem/utils.py'),
                              'onprem.utils.get_datadir': ('utils.html#get_datadir', 'onprem/utils.py'),
                              'onprem.utils.split_list': ('utils.html#split_list', 'onprem/utils.py')},
            'onprem.webapp': {}}}
